/* ============================================================================
 * Copyright (c) 2010, Michael A. Jackson (BlueQuartz Software)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright notice, this
 * list of conditions and the following disclaimer in the documentation and/or
 * other materials provided with the distribution.
 *
 * Neither the name of Michael A. Jackson nor the names of its contributors may
 * be used to endorse or promote products derived from this software without
 * specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 //
 //  This code was written under United States Air Force Contract number
 //                           FA8650-07-D-5800
 //
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

#include "CMUSurfaceMeshing2Filter.h"

#include "MXA/Utilities/MXADir.h"
#include "MXA/Utilities/MXAFileInfo.h"

#include "SurfaceMeshStructs.h"
#include "SurfaceMeshGeneralMCABoundedMulti.h"

// -----------------------------------------------------------------------------
//
// -----------------------------------------------------------------------------
CMUSurfaceMeshing2Filter::CMUSurfaceMeshing2Filter() :
AbstractFilter(),
m_GrainIdsArrayName(DREAM3D::CellData::GrainIds),
m_OutputNodesFile(""),
m_OutputEdgesFile(""),
m_OutputTrianglesFile(""),
m_XStep(1.0),
m_YStep(1.0),
m_ZStep(1.0),
m_GrainIds(NULL)
{
  setupFilterParameters();
}

// -----------------------------------------------------------------------------
//
// -----------------------------------------------------------------------------
CMUSurfaceMeshing2Filter::~CMUSurfaceMeshing2Filter()
{
}

// -----------------------------------------------------------------------------
//
// -----------------------------------------------------------------------------
void CMUSurfaceMeshing2Filter::setupFilterParameters()
{
  std::vector<FilterParameter::Pointer> parameters;
//  {
//    FilterOption::Pointer option = FilterOption::New();
//    option->setHumanLabel("Output Directory");
//    option->setPropertyName("OutputDirectory");
//    option->setWidgetType(FilterOption::OutputFileWidget);
//    option->setValueType("string");
//    parameters.push_back(option);
//  }
  {
    FilterParameter::Pointer option = FilterParameter::New();
    option->setHumanLabel("Output Nodes File");
    option->setPropertyName("OutputNodesFile");
    option->setWidgetType(FilterParameter::OutputFileWidget);
    option->setValueType("string");
    parameters.push_back(option);
  }
  {
    FilterParameter::Pointer option = FilterParameter::New();
    option->setHumanLabel("Output Edges File");
    option->setPropertyName("OutputEdgesFile");
    option->setWidgetType(FilterParameter::OutputFileWidget);
    option->setValueType("string");
    parameters.push_back(option);
  }
  {
    FilterParameter::Pointer option = FilterParameter::New();
    option->setHumanLabel("Output Triangles File");
    option->setPropertyName("OutputTrianglesFile");
    option->setWidgetType(FilterParameter::OutputFileWidget);
    option->setValueType("string");
    parameters.push_back(option);
  }
  {
    FilterParameter::Pointer option = FilterParameter::New();
    option->setHumanLabel("X Step");
    option->setPropertyName("XStep");
    option->setWidgetType(FilterParameter::DoubleWidget);
    option->setValueType("float");
    parameters.push_back(option);
  }
  {
    FilterParameter::Pointer option = FilterParameter::New();
    option->setHumanLabel("Y Step");
    option->setPropertyName("YStep");
    option->setWidgetType(FilterParameter::DoubleWidget);
    option->setValueType("float");
    parameters.push_back(option);
  }
  {
    FilterParameter::Pointer option = FilterParameter::New();
    option->setHumanLabel("Z Step");
    option->setPropertyName("ZStep");
    option->setWidgetType(FilterParameter::DoubleWidget);
    option->setValueType("float");
    parameters.push_back(option);
  }
  setFilterParameters(parameters);
}
// -----------------------------------------------------------------------------
void CMUSurfaceMeshing2Filter::writeFilterParameters(AbstractFilterParametersWriter* writer)
{
//  writer->writeValue("OutputDirectory", getOutputDirectory() );
  writer->writeValue("OutputNodesFile", getOutputNodesFile() );
  writer->writeValue("OutputEdgesFile", getOutputEdgesFile() );
  writer->writeValue("OutputTrianglesFile", getOutputTrianglesFile() );
  writer->writeValue("XStep", getXStep() );
  writer->writeValue("YStep", getYStep() );
  writer->writeValue("ZStep", getZStep() );
}

// -----------------------------------------------------------------------------
//
// -----------------------------------------------------------------------------
void CMUSurfaceMeshing2Filter::dataCheck(bool preflight, size_t voxels, size_t fields, size_t ensembles)
{
  setErrorCondition(0);
  std::stringstream ss;
  DataContainer* m = getDataContainer();

//  if (true == m_OutputDirectory.empty())
//  {
//    ss << getHumanLabel() << " needs the Output directory set";
//    setErrorCondition(-387);
//  }
  if(true == m_OutputNodesFile.empty())
  {
    addErrorMessage(getHumanLabel(), "The output Nodes file needs to be set", -380);
    setErrorCondition(-380);
  }
  if(true == m_OutputEdgesFile.empty())
  {
    addErrorMessage(getHumanLabel(), "The output Edges file needs to be set", -381);
    setErrorCondition(-381);
  }
  if(true == m_OutputTrianglesFile.empty())
  {
    addErrorMessage(getHumanLabel(), "The output Triangles file needs to be set", -382);
    setErrorCondition(-382);
  }

  GET_PREREQ_DATA(m, DREAM3D, CellData, GrainIds, ss, -300, int32_t, Int32ArrayType, voxels, 1);

}


// -----------------------------------------------------------------------------
//
// -----------------------------------------------------------------------------
void CMUSurfaceMeshing2Filter::preflight()
{
  /* Place code here that sanity checks input arrays and input values. Look at some
  * of the other DREAM3DLib/Filters/.cpp files for sample codes */
  dataCheck(true, 1, 1, 1);
}

// -----------------------------------------------------------------------------
//
// -----------------------------------------------------------------------------
void CMUSurfaceMeshing2Filter::execute()
{
  int err = 0;
  std::stringstream ss;
  setErrorCondition(err);
  DataContainer* m = getDataContainer();
  if(NULL == m)
  {
    setErrorCondition(-1);
    std::stringstream ss;
    ss << " DataContainer was NULL";
    PipelineMessage em(getNameOfClass(), ss.str(), -1);
    addErrorMessage(em);
    return;
  }
  setErrorCondition(0);
  int64_t totalPoints = m->getTotalPoints();
  size_t totalFields = m->getNumFieldTuples();
  size_t totalEnsembles = m->getNumEnsembleTuples();
  dataCheck(false, totalPoints, totalFields, totalEnsembles);
  if(getErrorCondition() < 0)
  {
    return;
  }
  size_t dims[3];
  m->getDimensions(dims);
  // Copy the Grain Ids into the "spin" component of each 'voxel' pointer
  voxel* point = (voxel *)malloc((totalPoints + 1) * sizeof(voxel));

  // Make sure all the directories are created for all the output files
  {
    std::string parentPath = MXAFileInfo::parentPath(getOutputNodesFile());
    if(!MXADir::mkdir(parentPath, true))
    {
      std::stringstream ss;
      ss << "Error creating parent path '" << parentPath << "'";
      notifyErrorMessage(ss.str(), -1);
      setErrorCondition(-1);
      return;
    }
  }
  {
    std::string parentPath = MXAFileInfo::parentPath(getOutputEdgesFile());
    if(!MXADir::mkdir(parentPath, true))
    {
      std::stringstream ss;
      ss << "Error creating parent path '" << parentPath << "'";
      notifyErrorMessage(ss.str(), -1);
      setErrorCondition(-1);
      return;
    }
  }
  {
    std::string parentPath = MXAFileInfo::parentPath(getOutputTrianglesFile());
    if(!MXADir::mkdir(parentPath, true))
    {
      std::stringstream ss;
      ss << "Error creating parent path '" << parentPath << "'";
      notifyErrorMessage(ss.str(), -1);
      setErrorCondition(-1);
      return;
    }
  }

  MeshParameters mp;
  mp.point = point;
  mp.xnum = static_cast<int>(dims[0]);
  mp.ynum = static_cast<int>(dims[1]);
  mp.znum = static_cast<int>(dims[2]);
  mp.xstep = static_cast<double>(getXStep());
  mp.ystep = static_cast<double>(getYStep());
  mp.zstep = static_cast<double>(getZStep());
  mp.nodes_file = getOutputNodesFile();
  mp.edges_file = getOutputEdgesFile();
  mp.triangles_file = getOutputTrianglesFile();
  mp.verbose = true;


  // Point 0 is a garbage...
  point[0].spin = 0;
  for (int64_t p = 0; p < totalPoints; ++p)
  {
    point[p + 1].spin = m_GrainIds[p];
  }
  double tx, ty, tz;
  size_t id = 0;
  // Let's fill out the coordinate of each voxel. Remember that x coordinate goes fastest...
  for (size_t k = 0; k < dims[2]; k++)
  {
    for (size_t j = 0; j < dims[1]; j++)
    {
      for (size_t i = 0; i < dims[0]; i++)
      {
        id = k * dims[0] * dims[1] + j * dims[0] + i + 1;
        tx = (double)(i) * mp.xstep;
        ty = (double)(j) * mp.ystep;
        tz = (double)(k) * mp.zstep;
        //printf("%10d %6.3f %6.3f %6.3f\n", id, tx, ty, tz);
        point[id].coord[0] = tx;
        point[id].coord[1] = ty;
        point[id].coord[2] = tz;
      }
    }
  }

  // Call the surface meshing function now
  err = CMUSurfaceMeshing2::execute(&mp);
  std::cout << getNameOfClass() << " Ending -----------------------------------------------" << std::endl;
  if(err < 0)
  {
    setErrorCondition(err);
    ss.str("");
    ss << " Error: Unknown error returned from Surface Meshing code";
    PipelineMessage em(getNameOfClass(), ss.str(), -1);
    addErrorMessage(em);
  }

  notifyStatusMessage("Complete");
}
