
/* ============================================================================
* Copyright (c) 2012 Dr. Sukbin Lee (Carnegie-Mellon University)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without modification,
* are permitted provided that the following conditions are met:
*
* Redistributions of source code must retain the above copyright notice, this
* list of conditions and the following disclaimer.
*
* Redistributions in binary form must reproduce the above copyright notice, this
* list of conditions and the following disclaimer in the documentation and/or
* other materials provided with the distribution.
*
* Neither the name of Anthony Rollet, Sukbin Lee, Jason Gruber, Stephen Sintay
* nor the names of its contributors may be used to endorse
* or promote products derived from this software without specific prior written
* permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
* FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
* DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
* SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
* CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
* OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
* USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */


#include <iostream>
#include <string>
#include <vector>

#include "Code/CMUSurfaceMeshing2Filters/SurfaceMeshStructs.h"
#include "Code/CMUSurfaceMeshing2Filters/SurfaceMeshGeneralMCABoundedMulti.h"
#include "Code/CMUSurfaceMeshing2Filters/Surface_Mesh_General_MCA_Bounded_Multi.cpp"
#include "Code/CMUSurfaceMeshing2Filters/DxReader.cpp"



void tokenize(const std::string& str, std::vector<std::string>& tokens, const std::string& delimiters = " ")
{
  // Skip delimiters at beginning.
  std::string::size_type lastPos = str.find_first_not_of(delimiters, 0);

  // Find first "non-delimiter".
  std::string::size_type pos = str.find_first_of(delimiters, lastPos);

  while (std::string::npos != pos || std::string::npos != lastPos)
  {
    // Found a token, add it to the vector.
    tokens.push_back(str.substr(lastPos, pos - lastPos));

    // Skip delimiters.  Note the "not_of"
    lastPos = str.find_first_not_of(delimiters, pos);

    // Find next "non-delimiter"
    pos = str.find_first_of(delimiters, lastPos);
  }
}



// Main function starts...
int main(int argc, char **argv)
{

  if (argc != 4)
  {
    std::cout << argv[0] << " needs 3 arguments:" << std::endl;
    std::cout << "  [1] The Input .dx file" << std::endl;
    std::cout << "  [2] Comma separated list of Step Sizes (0.25,0.25,0.25)" << std::endl;
    std::cout << "  [3] Output Base name for the Nodes, Edges, Triangles file to use." << std::endl;
    return EXIT_FAILURE;
  }


  std::string ifName(argv[1]); // Arg 1 is the Input File

  std::string stepSize(argv[2]); // Arg 2 is the step size as a comma delimited string
  std::vector<std::string> tokens;
  tokenize(stepSize, tokens, ",");
  double xstep = atof(tokens[0].c_str());
  double ystep = atof(tokens[1].c_str());
  double zstep = atof(tokens[2].c_str());

  std::string outputBaseName(argv[3]); // Arg 3 is the output file path

  voxel* point = NULL;     // contains voxel information...
//  int NS; // The number of sites(voxels) in the simulation box...
  int xnum, ynum, znum;

  std::cout << "Reading Microstructure...." << std::endl;
  point = initialize_micro_from_dx_file(xnum, ynum, znum, xstep, ystep, zstep, ifName);
  std::cout << "input file name = " << ifName << std::endl;
  std::cout << "step size in x, y, z = " << xstep << " " << ystep << " " << zstep << std::endl;
  std::cout << "dimensions of the microstructure = " << xnum << " " << ynum << " " << znum << std::endl;

  MeshParameters mp;
  mp.point = point;
  mp.xnum = xnum;
  mp.ynum = ynum;
  mp.znum = znum;
  mp.xstep = xstep;
  mp.ystep = ystep;
  mp.zstep = zstep;
  //mp.outputBaseName = outputBaseName;
  mp.nodes_file = outputBaseName + "_nodes.txt";
  mp.edges_file = outputBaseName + "_edges.txt";
  mp.triangles_file = outputBaseName + "_triangles.txt";
  mp.verbose = true;

  int err = execute(&mp);

  // Clean up the memory
  free(point);

  return err;
}

