/* ============================================================================
 * Copyright (c) 2012 Michael A. Jackson (BlueQuartz Software)
 * Copyright (c) 2012 Dr. Michael A. Groeber (US Air Force Research Laboratories)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright notice, this
 * list of conditions and the following disclaimer in the documentation and/or
 * other materials provided with the distribution.
 *
 * Neither the name of Michael A. Groeber, Michael A. Jackson, the US Air Force,
 * BlueQuartz Software nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior written
 * permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  This code was written under United States Air Force Contract number
 *                           FA8650-07-D-5800
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

#include "MFESurfaceSmoothingFilter.h"

#include "MFE_Smooth.h"

#ifndef MFE_SMOOTH_LIBRARY
#include "MFE_Smooth.cpp"
#endif
// -----------------------------------------------------------------------------
//
// -----------------------------------------------------------------------------
MFESurfaceSmoothingFilter::MFESurfaceSmoothingFilter() :
AbstractFilter(),
m_IterationSteps(0),
m_NodeConstraints(true),
m_ConstrainSurfaceNodes(true),
m_ConstrainQuadPoints(true),
m_SmoothTripleLines(true)
{
  setupFilterParameters();
}

// -----------------------------------------------------------------------------
//
// -----------------------------------------------------------------------------
MFESurfaceSmoothingFilter::~MFESurfaceSmoothingFilter()
{
}

// -----------------------------------------------------------------------------
//
// -----------------------------------------------------------------------------
void MFESurfaceSmoothingFilter::setupFilterParameters()
{
  std::vector<FilterParameter::Pointer> parameters;
 {
   FilterParameter::Pointer option = FilterParameter::New();
   option->setHumanLabel("Nodes File");
   option->setPropertyName("NodesFile");
   option->setWidgetType(FilterParameter::InputFileWidget);
   option->setValueType("string");
   parameters.push_back(option);
 }
 {
   FilterParameter::Pointer option = FilterParameter::New();
   option->setHumanLabel("Edges File");
   option->setPropertyName("EdgesFile");
   option->setWidgetType(FilterParameter::InputFileWidget);
   option->setValueType("string");
   parameters.push_back(option);
 }
 {
   FilterParameter::Pointer option = FilterParameter::New();
   option->setHumanLabel("Triangles File");
   option->setPropertyName("TrianglesFile");
   option->setWidgetType(FilterParameter::InputFileWidget);
   option->setValueType("string");
   parameters.push_back(option);
 }
 {
   FilterParameter::Pointer option = FilterParameter::New();
   option->setHumanLabel("Output Nodes File");
   option->setPropertyName("OutputNodesFile");
   option->setWidgetType(FilterParameter::OutputFileWidget);
   option->setValueType("string");
   parameters.push_back(option);
 }
 {
   FilterParameter::Pointer option = FilterParameter::New();
   option->setHumanLabel("Iteration Steps");
   option->setPropertyName("IterationSteps");
   option->setWidgetType(FilterParameter::IntWidget);
   option->setValueType("int");
   parameters.push_back(option);
 }
 {
   FilterParameter::Pointer option = FilterParameter::New();
   option->setHumanLabel("Apply Node Contraints");
   option->setPropertyName("NodeConstraints");
   option->setWidgetType(FilterParameter::BooleanWidget);
   option->setValueType("bool");
   parameters.push_back(option);
 }
 {
   FilterParameter::Pointer option = FilterParameter::New();
   option->setHumanLabel("Constrain Surface Nodes");
   option->setPropertyName("ConstrainSurfaceNodes");
   option->setWidgetType(FilterParameter::BooleanWidget);
   option->setValueType("bool");
   parameters.push_back(option);
 }
 {
   FilterParameter::Pointer option = FilterParameter::New();
   option->setHumanLabel("Constrain Quad Points");
   option->setPropertyName("ConstrainQuadPoints");
   option->setWidgetType(FilterParameter::BooleanWidget);
   option->setValueType("bool");
   parameters.push_back(option);
 }
 {
   FilterParameter::Pointer option = FilterParameter::New();
   option->setHumanLabel("Smooth Triple Lines");
   option->setPropertyName("SmoothTripleLines");
   option->setWidgetType(FilterParameter::BooleanWidget);
   option->setValueType("bool");
   parameters.push_back(option);
 }
  setFilterParameters(parameters);
}

// -----------------------------------------------------------------------------
//
// -----------------------------------------------------------------------------
void MFESurfaceSmoothingFilter::writeFilterParameters(AbstractFilterParametersWriter* writer)
{
  /* Place code that will write the inputs values into a file. reference the
  * AbstractFilterParametersWriter class for the proper API to use. */
    writer->writeValue("NodesFile", getNodesFile() );
    writer->writeValue("TrianglesFile", getTrianglesFile() );
    writer->writeValue("EdgesFile", getEdgesFile() );
    writer->writeValue("OutputNodesFile", getOutputNodesFile());
    writer->writeValue("IterationSteps", getIterationSteps());
    writer->writeValue("ConstrainNodes", getNodeConstraints());
    writer->writeValue("ConstrainSurfaceNodes", getConstrainSurfaceNodes());
    writer->writeValue("ConstrainQuadPoints", getConstrainQuadPoints());
    writer->writeValue("SmoothTripleLines", getSmoothTripleLines());
}

// -----------------------------------------------------------------------------
//
// -----------------------------------------------------------------------------
void MFESurfaceSmoothingFilter::dataCheck(bool preflight, size_t voxels, size_t fields, size_t ensembles)
{
  setErrorCondition(0);
  std::stringstream ss;

  if (m_NodesFile.empty() == true)
  {
    ss.str("");
    ss  << "The Nodes file must be set before executing this filter.";
    PipelineMessage em(getHumanLabel(), ss.str(), -1101);
    addErrorMessage(em);
    setErrorCondition(-1101);
  }
  if (m_TrianglesFile.empty() == true)
  {
    ss  << "The Triangles file must be set before executing this filter.";
    PipelineMessage em(getHumanLabel(), ss.str(), -1102);
    addErrorMessage(em);
    setErrorCondition(-1102);
  }
  if (m_EdgesFile.empty() == true)
  {
    ss  << "The Edges file must be set before executing this filter.";
    PipelineMessage em(getHumanLabel(), ss.str(), -1103);
    addErrorMessage(em);
    setErrorCondition(-1103);
  }
  if (m_OutputNodesFile.empty() == true)
  {
    ss  << "The Output Nodes file must be set before executing this filter.";
    PipelineMessage em(getHumanLabel(), ss.str(), -1104);
    addErrorMessage(em);
    setErrorCondition(-1104);
  }

}


// -----------------------------------------------------------------------------
//
// -----------------------------------------------------------------------------
void MFESurfaceSmoothingFilter::preflight()
{
  dataCheck(true, 1, 1, 1);
}

// -----------------------------------------------------------------------------
//
// -----------------------------------------------------------------------------
void MFESurfaceSmoothingFilter::setCancel(bool b)
{
  AbstractFilter::setCancel(b);
  if (b) {
    m_Smooth.cancel();
  }
}

// -----------------------------------------------------------------------------
//
// -----------------------------------------------------------------------------
void MFESurfaceSmoothingFilter::execute()
{
  int err = 0;
  setErrorCondition(err);
  DataContainer* m = getDataContainer();
  if(NULL == m)
  {
    setErrorCondition(-1);
    std::stringstream ss;
    ss << "DataContainer was NULL";
    PipelineMessage em(getHumanLabel(), ss.str(), -666);
    addErrorMessage(em);
    return;
  }
 // std::cout << getHumanLabel() << " Starting -----------------------------------------------" << std::endl;
  setErrorCondition(0);

  int error = m_Smooth.execute(m_NodesFile, m_TrianglesFile, m_EdgesFile, m_IterationSteps,
                        m_NodeConstraints, m_ConstrainSurfaceNodes, m_ConstrainQuadPoints,
                        m_SmoothTripleLines, m_RestartFile, m_OutputNodesFile);
  setErrorCondition(error);
//  std::cout << getHumanLabel() << " Ending -----------------------------------------------" << std::endl;

  notifyStatusMessage("Complete");
}
