
/* ============================================================================
 * Copyright (c) 2012 Dr. Anthony Rollet (Carnegie-Mellon University)
 * Copyright (c) 2012 Dr. Sukbin Lee (Carnegie-Mellon University)
 * Copyright (c) 2012 Dr. Jason Gruber (Carnegie-Mellon University)
 * Copyright (c) 2012 Dr. Stephen Sintay (Carnegie-Mellon University)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright notice, this
 * list of conditions and the following disclaimer in the documentation and/or
 * other materials provided with the distribution.
 *
 * Neither the name of Anthony Rollet, Sukbin Lee, Jason Gruber, Stephen Sintay
 * nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior written
 * permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
// smooth.cpp

//  g++ -g -O3 -o smooth smooth-grain3d-ADR.cpp

// Written by Jason Gruber jgruber@andrew.cmu.edu
//   many mods by Steve Sintay, ADR

//  added an option to turn off edge-of-model constraints
//  ADR, Nov 09; found another place to turn off dc/dt, Dec 09
//  ADR, Mar 10; added option to constrain quad points
//  ADR, May 10; added node info to output to track
//               problems with surface pull-in along GBs
//               changed the way in which constraints are added, and used
//  ADR, June 10; added edge info to compute TJ smoothing forces
//      created a new class for edges, but do not need it, in fact
//      created a function of line curvature for computing forces
//      also added option to restart from an INP
//       Feb 11; debugging surface node constraint

//  look for mobility values etc, and perhaps tweak them

#ifndef MFE_TRIANGLE_H_
#define MFE_TRIANGLE_H_

class triangle
{
  public:
    node& operator[](int i)
    {
      return *p[i];
    }
    const node& operator[](int i) const
    {
      return *p[i];
    }
    node*& operator()(int i)
    {
      return p[i];
    }
    node* const & operator()(int i) const
    {
      return p[i];
    }
    double area() const;
    double aspect() const;
    double MinDihedral() const;
    double circularity(double area) const;
    Vector<double> normal() const;
    int region1;
    int region2;

  private:
    node* p[3];
};

double triangle::area() const
{
  double a[3], b[3], c[3];
  a[0] = (*p[1])[0] - (*p[0])[0];
  a[1] = (*p[1])[1] - (*p[0])[1];
  a[2] = (*p[1])[2] - (*p[0])[2];
  b[0] = (*p[2])[0] - (*p[0])[0];
  b[1] = (*p[2])[1] - (*p[0])[1];
  b[2] = (*p[2])[2] - (*p[0])[2];
  c[0] = a[1] * b[2] - a[2] * b[1];
  c[1] = a[2] * b[0] - a[0] * b[2];
  c[2] = a[0] * b[1] - a[1] * b[0];
  return 0.5 * sqrt(c[0] * c[0] + c[1] * c[1] + c[2] * c[2]);
}

Vector<double> triangle::normal() const
{
  Vector<double> n(3);
  double a[3], b[3];
  a[0] = (*p[1])[0] - (*p[0])[0];
  a[1] = (*p[1])[1] - (*p[0])[1];
  a[2] = (*p[1])[2] - (*p[0])[2];
  b[0] = (*p[2])[0] - (*p[0])[0];
  b[1] = (*p[2])[1] - (*p[0])[1];
  b[2] = (*p[2])[2] - (*p[0])[2];
  n[0] = a[1] * b[2] - a[2] * b[1];
  n[1] = a[2] * b[0] - a[0] * b[2];
  n[2] = a[0] * b[1] - a[1] * b[0];
  double norm = sqrt(n[0] * n[0] + n[1] * n[1] + n[2] * n[2]);
  if(norm > 0.0)
  {
    double rnorm = 1.0 / norm;
    n[0] *= rnorm;
    n[1] *= rnorm;
    n[2] *= rnorm;
  }
  return n;
}

double triangle::aspect() const
{
  double a = distance(*p[0], *p[1]);
  double b = distance(*p[1], *p[2]);
  double c = distance(*p[2], *p[0]);
  double min = a;
  double max = a;
  if(b > max) max = b;
  else if(b < min) min = b;
  if(c > max) max = c;
  else if(c < min) min = c;
  return max / min;
}

double triangle::circularity(double area) const
{
  double a = distance(*p[0], *p[1]);
  double b = distance(*p[1], *p[2]);
  double c = distance(*p[2], *p[0]);
  double s = 0.5 * (a + b + c); //  1/2 perimeter
  double r = area / s;
  double R = a * b * c / 4 / area;
  return R / r;
}

double triangle::MinDihedral() const //  another triangle quality measure
{
  //  a points from p0 to p1; b is p0 to p2;  c points p1 to p2
  double a[3], b[3], c[3], min, w1, w2, w3;
  a[0] = (*p[1])[0] - (*p[0])[0];
  a[1] = (*p[1])[1] - (*p[0])[1];
  a[2] = (*p[1])[2] - (*p[0])[2];
  double norm = sqrt(a[0] * a[0] + a[1] * a[1] + a[2] * a[2]);
  if(norm > 0.0)
  {
    double rnorm = 1.0 / norm;
    a[0] *= rnorm;
    a[1] *= rnorm;
    a[2] *= rnorm;
  }
  //  std::cout << "MinDIhedral, a: " << a[0] <<" " << a[1] <<" " << a[2] << std::endl ;
  b[0] = (*p[2])[0] - (*p[0])[0];
  b[1] = (*p[2])[1] - (*p[0])[1];
  b[2] = (*p[2])[2] - (*p[0])[2];
  norm = sqrt(b[0] * b[0] + b[1] * b[1] + b[2] * b[2]);
  if(norm > 0.0)
  {
    double rnorm = 1.0 / norm;
    b[0] *= rnorm;
    b[1] *= rnorm;
    b[2] *= rnorm;
  }
  //  std::cout << "MinDIhedral, b: " << b[0] <<" " << b[1] <<" " << b[2] << std::endl ;
  c[0] = (*p[2])[0] - (*p[1])[0];
  c[1] = (*p[2])[1] - (*p[1])[1];
  c[2] = (*p[2])[2] - (*p[1])[2];
  norm = sqrt(c[0] * c[0] + c[1] * c[1] + c[2] * c[2]);
  if(norm > 0.0)
  {
    double rnorm = 1.0 / norm;
    c[0] *= rnorm;
    c[1] *= rnorm;
    c[2] *= rnorm;
  }
  //  std::cout << "MinDIhedral, c: " << c[0] <<" " << c[1] <<" " << c[2] << std::endl ;
  w1 = a[0] * b[0] + a[1] * b[1] + a[2] * b[2];
  if(w1 < -1.) w1 = -1.;
  if(w1 > 1.) w1 = 1.;
  w2 = -1. * a[0] * c[0] - a[1] * c[1] - a[2] * c[2];
  if(w2 < -1.) w2 = -1.;
  if(w2 > 1.) w2 = 1.;
  w3 = c[0] * b[0] + c[1] * b[1] + c[2] * b[2];
  if(w3 < -1.) w3 = -1.;
  if(w3 > 1.) w3 = 1.;
  //  std::cout << "MinDIhedral, w: " << w1 <<" " << w2 <<" " << w3 << std::endl ;
  double d1 = acos(w1);
  double d2 = acos(w2);
  double d3 = acos(w3);
  // debug
  //  std::cout << "MinDIhedral, angles: " << d1*180./PI <<" " << d2*180./PI <<" " << d3*180./PI << std::endl ;
  min = d1;
  if(d2 < min) min = d2;
  if(d3 < min) min = d3;
  return min;
}

#endif /* MFE_TRIANGLE_H_ */
