
/* ============================================================================
 * Copyright (c) 2012 Dr. Anthony Rollet (Carnegie-Mellon University)
 * Copyright (c) 2012 Dr. Sukbin Lee (Carnegie-Mellon University)
 * Copyright (c) 2012 Dr. Jason Gruber (Carnegie-Mellon University)
 * Copyright (c) 2012 Dr. Stephen Sintay (Carnegie-Mellon University)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright notice, this
 * list of conditions and the following disclaimer in the documentation and/or
 * other materials provided with the distribution.
 *
 * Neither the name of Anthony Rollet, Sukbin Lee, Jason Gruber, Stephen Sintay
 * nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior written
 * permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
// smooth.cpp

//  g++ -g -O3 -o smooth smooth.cpp

// Written by Jason Gruber jgruber@andrew.cmu.edu
//   many mods by Steve Sintay, ADR

//  added an option to turn off edge-of-model constraints
//  ADR, Nov 09; found another place to turn off dc/dt, Dec 09
//  ADR, Mar 10; added option to constrain quad points
//  ADR, May 10; added node info to output to track
//               problems with surface pull-in along GBs
//               changed the way in which constraints are added, and used
//  ADR, June 10; added edge info to compute TJ smoothing forces
//      created a new class for edges, but do not need it, in fact
//      created a function of line curvature for computing forces
//      also added option to restart from an INP
//       Feb 11; debugging surface node constraint

//  look for mobility values etc, and perhaps tweak them

#include "MFE_Smooth.cpp"

#include <iostream>

int main(int argc, char* argv[])
{

  std::cout << "Welcome to Smooth-Grain3D " << std::endl;
  std::cout << "Originally by Jason Gruber, modified by Steve Sintay, Tony Rollett " << std::endl;

  std::cout << "num. args = " << argc - 1 << std::endl;
  if(argc - 1 < 8)
  {
    std::cout << "expecting 8 arguments, as follows" << std::endl;
    std::cout
        << "Usage: ./smooth nodes.txt triangles.txt output.inp steps constrain-all[1=no] surface-constrain[1=no] quad-constrain[1=no] SmoothTripleLines[1=no]"
        << std::endl;
    ;
    std::cout << "- surface-constrain (anything other than 1) means limiting surface nodes to stay in their surface plane" << std::endl;
    std::cout << "- constrain-all (anything other than 1) means locking surface nodes to stay put (no degrees of freedom)" << std::endl;
    std::cout << "- quad-constrain (anything other than 1) means locking quad points in place" << std::endl;
    std::cout << "- smooth-triple-lines (anything other than 1) means allowing triple lines to be smoothed" << std::endl;
    std::cout << "if an additional argument with a XXX.inp file is present, re-start will be assumed" << std::endl;
    std::cout << "nodes.txt will be read in (for type etc.) but node position will be taken from the inp file" << std::endl;
    //  std::cout << "  then you can specify the starting iteration for the output file"<< std::endl;
    //  std::cout << std::endl;
    //  std::cout << "./smooth  output 34:200" << std::endl;
    //  std::cout << std::endl;
    //  std::cout << " will start the output file iterations at 34 and end at 200." << std::endl;
    exit(1);
  }

  std::string nodesFile(argv[1]);
  std::string trianglesFile(argv[2]);

  // extract the basename from the provided filename
  std::string infile(argv[3]);
  std::vector<std::string> tokens;
  std::string delimiters = "."; // Only a period
  std::string basename;
  tokenize(infile, tokens, delimiters);

  basename = tokens.front();
  std::cout << " Using: " << basename << " As basename " << std::endl;
  if(tokens.size() > 2)
  {
    std::cout << "WARNING: multiple \".\" in file name: " << infile << std::endl;
  }

  // output sukbin nodes file
  std::string outputNodesFile = basename + "_nodes_MFE_smoothed.txt";

  // read in nodes/triangles
  std::ifstream input1(nodesFile.c_str());
  std::ifstream input2(trianglesFile.c_str());
  std::ofstream velocities;

  int nnod, ntri;

  input1 >> nnod;
  input2 >> ntri;
  input1.close();
  input2.close();

  std::cout << "Number of nodes: " << nnod << std::endl;
  std::cout << "Number of triangles: " << ntri << std::endl;
  std::cout << std::endl;

  int steps = atoi(argv[4]);

  int constrain;
  constrain = 1;
  if(atoi(argv[5]) == 1)
  {
    constrain = 0;
    std::cout << "node constraints will NOT be applied" << std::endl;
  }
  else
  {
    std::cout << "node constraints WILL be applied" << std::endl;
  }

  int constrainSurface;
  constrainSurface = 1;
  if(atoi(argv[6]) == 1)
  {
    constrainSurface = 0;
    std::cout << "Surface nodes will NOT be totally constrained" << std::endl;
  }
  else
  {
    std::cout << "Surface nodes WILL be totally constrained" << std::endl;
  }

  int constrainQuads;
  constrainQuads = 1;
  if(atoi(argv[7]) == 1)
  {
    constrainQuads = 0;
    std::cout << "Quad point constraints will NOT be applied" << std::endl;
  }
  else
  {
    std::cout << "Quad point constraints WILL be applied" << std::endl;
  }

  int smoothTripleLines;
  smoothTripleLines = 1;
  if(atoi(argv[8]) == 1)
  {
    smoothTripleLines = 0;
    std::cout << "Triple lines will NOT be smoothed" << std::endl;
  }
  else
  {
    std::cout << "Triple Lines WILL be smoothed;" << std::endl;
  }

  std::string restartFile;
  if(argc - 1 >= 9)
  {
    restartFile = argv[9];
  }

  std::string edgesFile("edges.txt");
  MFESmooth smooth;
  return smooth.execute(nodesFile, trianglesFile, edgesFile, steps, constrain, constrainSurface, constrainQuads, smoothTripleLines, restartFile, outputNodesFile);

}

